/**
 * @author [Name of the original author of the file goes here]
 * @file [path from intel-iscsi-v20-3.0/src]
 * 
 * [Description of the contents of the goes here.  The first line serves as
 * the "brief description and the entire description serves as
 * the "detailed description.]
 */


#ifndef __CONTROLLER_REPLICATION_RAID_C
#define __CONTROLLER_REPLICATION_RAID_C


/* Header files */
#include <scsi/scsi.h>
#include <sys/types.h>
#include <sys/uio.h>

#include "iscsi.h"
#include "device.h"
#include "controller_replication_raid.h"
#include "initiator.h"
#include "assert.h"
#include "tests.h"

/* ****************** Defines *******************************************/


/* ****************** Enumeration declarations **************************/


/* ****************** Struct declarations *******************************/


/* ****************** Typedefs ******************************************/


/* ****************** Global variable declaraitons **********************/

/** Tells whether to perform replication or RAID */
controller_type_t g_controller_type = replication;

/**
 *The number of blocks seen by the controller.  This is the sum of the number
 * of blocks exported by each target
 */
extern uint64_t disk_num_blocks;

/**
 * The block length of each block.  This should be the same across all targets */
extern unsigned disk_block_len;

/** The number of targets */
unsigned long long g_num_targets = 0;


/* ****************** Macros ********************************************/


/* ****************** Private function declarations *********************/


/* ****************** Private functions *********************************/


/* ****************** Public functions **********************************/

uint64_t controller_read( ISCSI_SCSI_CMD_T *args, uint64_t lba, unsigned short len, void *ptr )
{
    /*
     * Fill in this function properly
     * The following example serves as a passthrough to one udisk
     */
    struct iovec vector[1];
    
    memset( ptr, 0, len * disk_block_len );
    vector[0].iov_base = ptr;
    vector[0].iov_len = len*disk_block_len;
    
    read_or_write( 0, 0, lba, len, disk_block_len,
                   (unsigned char *)vector, 1, 0, 1);

    /* Change this to return the actual number of blocks read */
    return( len * disk_block_len ); 
}


uint64_t  controller_write( TARGET_CMD_T *cmd, uint64_t lba, unsigned short len,
                             const struct iovec *vector, size_t count)
{
    /*
     * Fill in this function properly
     * The following example serves as a pass through to the first udisk
     */
    read_or_write( 0, 0, lba, len, disk_block_len,
               (unsigned char *)vector, count, 1, 1);
     
     /* Change this to return the actual number of blocks read */
     return( len * disk_block_len );
}


int controller_init( char *target_locations )
{
    unsigned temp_max_lba = 0;
    unsigned temp_block_len = 0;
    unsigned old_block_len = 0;
    int i;
    
    g_num_targets = initiator_init(target_locations, 0);
    if( g_num_targets == -1 ) {
        return( -1 );
    }
    PRINT("Found %llu TARGETS OK\n", g_num_targets);
    assert( g_num_targets >= 1);
    
    /*
     * Issue READ_CAPACITY calls to determine block_len and num_blocks of
     * targets.  To make things easier, you should assert that block_len is the
     * same across all targets.
     */
    disk_num_blocks = 0;
    for( i = 0; i < g_num_targets; i++ ) {

        read_capacity( i, 0, &temp_max_lba, &temp_block_len, 1 );
        
        if( (i != 0) && (temp_block_len != old_block_len) ) {
            TRACE_ERROR("Target block sizes don't match");
            return(-1);
        }

        disk_num_blocks = (temp_max_lba+1);
        old_block_len = temp_block_len;
    }
    disk_block_len = temp_block_len;

    PRINT("Controller sees %llu blocks and a block size of %u",
          disk_num_blocks, disk_block_len );

    return( 0 );
}


void controller_set_type( controller_type_t type )
{
    assert( type == raid5 || type == replication );
    g_controller_type = type;
}

#endif /* __CONTROLLER_REPLICATION_RAID_C */
