/**
 * @author [Name of the original author of the file goes here]
 * @file [path from instel-iscsci-v20-3.0/src]
 * 
 * [Description of the contents of the goes here.  The first line
 * serves as the "brief description and the entire description serves
 * as the "detailed description.]
 */ 


#ifndef __CONTROLLER_REPLICATION_RAID_H
#define __CONTROLLER_REPLICATION_RAID_H


/* Header files */
#include "iscsi.h"
#include "device.h"


/* ****************** Defines *******************************************/ 


/* ****************** Enumeration declarations **************************/ 

/**
 * This defines whether or not the iSCSI controller should perform RAID or
 * replication
 */
typedef enum controller_type_e {
    replication,
    raid5
} controller_type_t;


/* ****************** Struct declarations *******************************/ 


/* ****************** Typedefs ******************************************/


/* ****************** Macros ********************************************/


/* ****************** Public function declarations **********************/

/**
 * @author [Insert author here]
 *
 * Function that reads data from files that are either striped or replicated
 * across num_device targets.
 *
 * @param args: This contains the reads arguments. For example, the CDB, etc.
 * @param lba: The logical block address at which to perform the read
 * @param len: The number of blocks to read
 *
 * @return: The actual number of bytes read
 */
uint64_t controller_read( ISCSI_SCSI_CMD_T *cmd, uint64_t lba, unsigned short len, void *ptr );


 /**
 * @author [Insert author here]
 *
 * Function that writes data from files that are either striped or replicated
 * across num_targets targets
 *
 * @param cmd: This contains the write arguments.  For example, the CDB, etc.
 * @param lba: The logical block address at which to perform the read
 * @param len: The number of blocks to read
 * @param vector: Scatter gather buffers containing the I/O to be written
 * @param len: The number of buffers in the scatter gather argument
 *
 * @return: The actual number of bytes read
 */
uint64_t controller_write( TARGET_CMD_T *cmd, uint64_t lba, unsigned short len,
                            const struct iovec *vector, size_t count );


/**
 * @author [Insert author here]
 * Initializes controller by initializing all targets.  The locations of targets
 * are specified by the filename in target_locations
 *
 * @param target_locations: IP addresses of the targets
 *
 * @return: 0 on success, -1 on failure
 */
int controller_init( char *target_locatoins );


/**
 * @author [Insert author here]
 *
 * Function that tells the iSCSI controller whether it should use RAID-5
 * semantics or Replication semantics
 *
 * @param type: raid5 for RAID-5, replication for Replication
 */
void controller_set_type( controller_type_t type );

     
#endif /* __CONTROLLER_REPLICATION_RAID_H */
